/******************************************************************************/
/* (c) 2006 jamie twycross, jpt AT cs.nott.ac.uk                              */
/* released under gnu gpl v2                                                  */
/* tcreplay - replay logfiles                                                 */
/******************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <unistd.h>

#include <stdio.h>
#include <sys/poll.h>
#include <err.h>

#include <tissue/tissue.h>
#include <tissue/syscall.h>

#define MAX_DS_ANTIGEN		100
#define MAX_DS_SIGNALS		100
#define MAX_LINE			1024

static char logfile[MAX_LINE];  /* log to replay */
static short debug = 0;  /* run in debug mode */
static char tshostname[MAX_LINE] = "localhost";
static int16_t tsport = TS_DEFAULT_PORT;
static short realtime = 1;  /* replay at realtime */
static char clientname[256] = "tcreplay";  /* name of client */
static short syscall_input = 0;  /* convert syscall names to numbers */

static void parse_command_line(int argc, char **argv)
{
	int c;

	while((c = getopt(argc, argv, "dfsp:h:")) != -1)
		switch(c) {
			case 'd':  /* debug - log to stdout */
				debug = 1;
				break;

			case 'f':  /* replay as fast as possible */
				realtime = 0;
				break;

			case 's':  /* syscall input - convert names to numbers */
				syscall_input = 1;
				break;

			case 'p':  /* port tissue server on */
				tsport = (int16_t) atoi(optarg);
				break;

			case 'h':  /* tissue server hostname */
				memset(tshostname, 0, sizeof(tshostname));
				strncpy(tshostname, optarg, sizeof(tshostname) - 1);
				break;

			default:
			usage:
				fprintf(stderr, \
						"usage: tcreplay " \
						"[-d -f -s -p PORT -h HOSTNAME] LOGFILE\n" \
						"\t-d - debug (print to stdout)\n" \
						"\t-f - replay as fast as possible\n" \
						"\t-s - syscall input\n" \
						"\t-p - tissue server port\n" \
						"\t-h - tissue server hostname\n" \
						"\tLOGFILE - log to replay\n");
				exit(EXIT_FAILURE);
				break;
		}

	if(optind != (argc - 1))
		goto usage;
	memset(logfile, 0, sizeof(logfile));
	strncpy(logfile, argv[optind], sizeof(logfile) - 1);
}

int main(int argc, char **argv)
{
	FILE *fp;
	char line[MAX_LINE], antigen[MAX_LINE], type[MAX_LINE], *ptr;
	struct timeval ts, sleeptime, lasttime;
	DS *dssignal, *dsantigen;
	Network_Info ts_net_info; /* network info on tissue server */
	unsigned int i;
	int id, num_antigen, num_signals, sc_number, count;
	double signal;

	parse_command_line(argc, argv);

	/* connect to tissue server */
	strcpy(ts_net_info.hostname, tshostname);
	ts_net_info.port = (int16_t) tsport;
	strcpy(ts_net_info.clientname, clientname);
	connect_ts(&ts_net_info);

	/* initialise signal datastream structure */
	dssignal = init_ds(MAX_DS_SIGNALS, 0);  /* 0 antigen */
	for(i = 0; i < MAX_DS_SIGNALS; i++) {
		dssignal->signal[i] = init_signal();
		dssignal->signal[i]->id = i;
	}

	/* initialise antigen datastream structure */
	dsantigen = init_ds(0, MAX_DS_ANTIGEN);  /* 0 signals */

	/* open log file */
	if(!(fp = fopen(logfile, "r"))) {
		disconnect_ts();
		err(1, "error opening log file");
	}

	/* send status message to server */
	snprintf(line, sizeof(line) - 1, "replay client replaying file %s", \
			logfile);
	send_message(line);

	/* get first timestamp */
	if(fgets(line, sizeof(line) - 1, fp)) {
		/* parse line */
		if(sscanf(line, "%u.%u ", (unsigned int *) &lasttime.tv_sec, \
				(unsigned int *) &lasttime.tv_usec) != 2)
			goto error;
		rewind(fp);
	}

	/* replay input from logfile */
	while(fgets(line, sizeof(line), fp)) {
		/* parse line */
		ptr = line;

		/* timestamp and type */
		if(sscanf(ptr, "%u.%u %s %n", (unsigned int *) &ts.tv_sec, \
				(unsigned int *) &ts.tv_usec, type, &count) != 3)
			goto error;
		ptr += count;

		/* work out time to sleep */
		timeval_diff(&lasttime, &ts, &sleeptime);

		if(strcmp(type, "antigen") == 0) {
			/* antigen */
			if(sscanf(ptr, "%d %d %n", &id, &num_antigen, &count) != 2) {
				fprintf(stderr, "error: failed to parse " \
						"line:\n%s\n", line);
				break;
			}
			ptr += count;
			if(num_antigen <= 0 || num_antigen > MAX_DS_ANTIGEN)
				goto error;
			dsantigen->num_antigen = num_antigen;
			/* set antigen timestamp */
			memcpy(&dsantigen->tm, &ts, sizeof(struct timeval));
			/* set antigen id */
			dsantigen->id = id;
			for(i = 0; i < num_antigen; i++) {
				if(sscanf(ptr, "%s %n", antigen, &count) != 1)
					goto error;
				if(syscall_input) {
					if((sc_number = syscall_number(antigen)) == -1)
						goto error;
					snprintf(antigen, sizeof(antigen), "%d", sc_number);
				}
				/* set antigen value */
				dsantigen->antigen[i] = \
						set_antigen(dsantigen->antigen[i], antigen);
				/* same id as datastream */
				dsantigen->antigen[i]->id = id;
				ptr += count;
				if(debug)
					print_ds(dsantigen);
				/* send to tissue server */
				send_ds(dsantigen);
			}
		} else {
			if(strcmp(type, "signal") == 0) {
				/* signal */
				if(sscanf(ptr, "%d %d %n", &id, &num_signals, &count) != 2)
					goto error;
				ptr += count;
				if(num_signals <= 0 || num_signals > MAX_DS_SIGNALS)
					goto error;
				dssignal->num_signals = num_signals;
				/* set signal timestamp */
				memcpy(&dssignal->tm, &ts, sizeof(struct timeval));
				/* set signal id */
				dssignal->id = id;
				for(i = 0; i < num_signals; i++) {
					if(*ptr == '-') {
						/* ignore signal value */
						dssignal->signal[i]->id = -1;
						dssignal->signal[i]->val = 0.0;
						count = 2;
					}
					else {
						if(sscanf(ptr, "%lf %n", &signal, &count) != 1)
							goto error;
						/* set signal value */
						dssignal->signal[i]->val = signal;
					}
					ptr += count;
				}
				if(debug)
					print_ds(dssignal);
				/* send to tissue server */
				send_ds(dssignal);
			} else 
				goto error;
		}

		/* sleep for time difference */
		memcpy(&lasttime, &ts, sizeof(struct timeval));
		if(realtime && !zerotime(sleeptime))
			count = select(0, 0, 0, 0, &sleeptime);
	}

	/* tidy up */
	disconnect_ts();  /* disconnect from tissue server */
	fclose(fp);
	return 0;

error:
	fprintf(stderr, "error: failed to parse line: %s", line);
	disconnect_ts();
	fclose(fp);
	return 1;
}
